# frozen_string_literal: true

require 'subconv/utility'

require 'solid_struct'
require 'timecode'

module Subconv
  module Scc
    FILE_MAGIC = 'Scenarist_SCC V1.0'

    # Grid size
    GRID_ROWS = 15
    GRID_COLUMNS = 32

    # Grid is just an array with some extra convenience functions and a default size
    class Grid < Array
      def initialize(*args)
        if args.empty?
          super(GRID_ROWS) { Array.new(GRID_COLUMNS) }
        elsif args.size == 1
          fail ArgumentError, 'Attempted to create grid from unsupported type' unless args[0].is_a?(Array)
          fail ArgumentError, 'Grid has illegal row count' if args[0].size != GRID_ROWS
          fail ArgumentError, 'Grid has illegal column count' if args[0].any? { |row| row.size != GRID_COLUMNS }

          super(args.first)
        else
          fail ArgumentError, 'Illegal number of parameters'
        end
      end

      # The grid is empty when there are no characters in it
      def empty?
        flatten.compact.empty?
      end

      # Make a deep copy
      def clone
        Grid.new(map { |row| row.map(&:clone) })
      end

      # Insert continuous text at a given position
      # Returns self for chaining
      def insert_text(row, column, text, style = CharacterStyle.default)
        text.each_char do |char|
          self[row][column] = Character.new(char, style)
          column += 1
        end
        self
      end

      def without_identical_characters(other_grid)
        result = clone
        each_with_index do |row, row_i|
          row.each_with_index do |column, column_i|
            result[row_i][column_i] = nil if other_grid[row_i][column_i] == column
          end
        end

        result
      end

      def to_simple_text
        map { |row| row.map { |char| char.nil? ? ' ' : char.character }.join }.join("\n")
      end
    end

    # Color constants as immutable value objects with some convenience functions (e.g. conversion to string or symbol)
    # All available colors are registered as constants in this class, e.g. Color::WHITE, Color::RED and so on
    # The instances of this class are all frozen and can never be changed.
    # Instances can be retrieved only by the constants or with ::for_value
    class Color
      def initialize(color)
        @color = color
      end

      # CEA-608 color code
      def value
        @color
      end
      alias to_i value

      # Lower-case CEA-608 name of the color
      def to_s
        to_symbol.to_s
      end
      alias inspect to_s

      # rubocop:disable MutableConstant
      COLORS = {}
      TO_SYMBOL_MAP = {}
      # rubocop:enable MutableConstant

      def self.register_color(name, value)
        # Make sure the new color is immutable
        new_color = Color.new(value).freeze
        # Register in lookup tables
        COLORS[value] = new_color
        TO_SYMBOL_MAP[value] = name
        # Register as class constant
        const_set(name.to_s.upcase, new_color)
      end

      # CEA-608 colors
      register_color :white, 0
      register_color :green, 1
      register_color :blue, 2
      register_color :cyan, 3
      register_color :red, 4
      register_color :yellow, 5
      register_color :magenta, 6

      # Prevent future modifications
      COLORS.freeze
      TO_SYMBOL_MAP.freeze

      # Lower-case CEA608 name of the color as symbol
      def to_symbol
        TO_SYMBOL_MAP[@color]
      end

      # Get the Color instance corresponding to a CEA608 color code
      def self.for_value(value)
        color = COLORS[value]
        fail "Color value #{value} is unknown" if color.nil?

        color
      end

      # Disallow creating new instances
      private_class_method :new
    end

    # Encapsulates properties of single characters
    CharacterStyle = SolidStruct.new(:color, :italics, :underline, :flash)
    class << CharacterStyle
      def default
        CharacterStyle.new(color: Color::WHITE, italics: false, underline: false, flash: false)
      end
    end

    # One character in the closed caption grid
    Character = SolidStruct.new(:character, :style)

    # One fully rendered caption displayed at a specific point in time
    class Caption
      def initialize(args)
        # Inject default
        args = { mode: :pop_on, char_replacement: false }.merge(args)
        self.timecode = args[:timecode]
        self.grid = args[:grid]
        self.mode = args[:mode]
        self.char_replacement = args[:char_replacement]
      end

      def ==(other)
        timecode == other.timecode && grid == other.grid && mode == other.mode && char_replacement? == other.char_replacement?
      end

      alias eql? ==

      attr_reader :mode
      attr_writer :char_replacement
      attr_accessor :timecode, :grid

      def mode=(mode)
        fail 'Unknown mode' unless %i[pop_on paint_on].include?(mode)

        @mode = mode
      end

      def pop_on_mode?
        @mode == :pop_on
      end

      def paint_on_mode?
        @mode == :paint_on
      end

      # Is this caption just replacing an existing character with an extended character?
      def char_replacement?
        @char_replacement
      end
    end

    # SCC reader
    # Parse and render an SCC file sequentially into a background and foreground grid
    # like a TV set would do and store the resulting closed captions as grid snapshots into an array
    # whenever the foreground grid changes.
    #
    # Only captions in data channel 1 are read. Also, invalid byte parity will raise an error unless checking is disabled.
    # The advanced recovery methods mentioned in CEA608 are not implemented since the source is assumed to contain no errors (e.g. DVD source).
    class Reader
      # Regular expression for parsing one line of data
      LINE_REGEXP = /^(?<timecode>[0-9:;]+)\t(?<data>(?:[0-9a-fA-F]{4} ?)+)$/.freeze

      # rubocop:disable MutableConstant

      # Map of standard characters that do not match the standard ASCII codes
      # to their corresponding unicode characters
      STANDARD_CHARACTER_MAP = {
        '\''   => '’', # CEA-608-B says to use curly apostrophe
        '*'    => 'á',
        '\\'   => 'é',
        '^'    =>	'í',
        '_'    =>	'ó',
        '`'    =>	'ú',
        '{'    =>	'ç',
        '|'    =>	'÷',
        '}'    =>	'Ñ',
        '~'    =>	'ñ',
        "\x7f" =>	'█'
      }
      # rubocop:enable MutableConstant
      # Simply return the character if no exception matched
      STANDARD_CHARACTER_MAP.default_proc = proc do |_hash, key|
        key
      end
      STANDARD_CHARACTER_MAP.freeze

      # Map of special characters to unicode codepoints
      # 0x39 transparent space is handled specially since it is not a real character
      SPECIAL_CHARACTER_MAP = Hash[%w[® ° ½ ¿ ™ ¢ £ ♪ à _ è á ê î ô û].each_with_index.map { |char, i| [0x30 + i, char] }].freeze

      # Extended character maps for Western European languages and box drawing
      EXTENDED_CHARACTER_MAPS = [
        Hash[%w[Á É Ó Ú Ü ü ‘ ¡ * ' ─ © ℠ · “ ” À Â Ç È Ê Ë ë Î Ï ï Ô Ù ù Û « »].each_with_index.map { |char, i| [0x20 + i, char] }].freeze,
        Hash[%w[Ã ã Í Ì ì Ò ò Õ õ { } \\ ^ _ | ~ Ä ä Ö ö ß ¥ ¤ │ Å å Ø ø ┌ ┐ └ ┘].each_with_index.map { |char, i| [0x20 + i, char] }].freeze
      ].freeze

      # Map of preamble address code high bytes to their
      # corresponding base row numbers (counted from 0)
      PREAMBLE_ADDRESS_CODE_ROW_MAP = {
        0x10 => 10,
        0x11 => 0,
        0x12 => 2,
        0x13 => 11,
        0x14 => 13,
        0x15 => 4,
        0x16 => 6,
        0x17 => 8
      }.freeze

      # Error classes
      class Error < RuntimeError; end
      class InvalidFormatError < Error; end
      class ParityError < Error; end

      # Internal state of the parser consisting of current drawing position, character style and captioning mode
      class State
        def initialize(params)
          self.row = params[:row]
          self.column = params[:column]
          @style = params[:style]
          self.mode = params[:mode]
        end

        attr_accessor :style
        attr_reader :row, :column, :mode, :char_replaced

        # Make sure the maximum row count is not exceeded
        def row=(row)
          @row = Utility.clamp(row, 0, GRID_ROWS - 1)
        end

        # Make sure the cursor does not get outside the screen left or right
        def column=(column)
          @column = Utility.clamp(column, 0, GRID_COLUMNS - 1)
        end

        def mode=(mode)
          fail 'Unknown mode' unless %i[pop_on paint_on].include?(mode)

          @mode = mode
        end

        def pop_on_mode?
          @mode == :pop_on
        end

        def paint_on_mode?
          @mode == :paint_on
        end

        def char_replaced=(replaced)
          fail ArgumentError, 'Invalid value for char_replaced' unless [true, false].include?(replaced)

          @char_replaced = replaced
        end

        def start_new_frame
          self.char_replaced = false
        end

        def self.default
          State.new(row: 0, column: 0, style: CharacterStyle.default, mode: :pop_on)
        end
      end

      # Actual conversion result
      attr_reader :captions

      # Read an SCC file from the IO object io for a video
      def read(io, fps, check_parity = true)
        # Initialize new grids for character storage
        @foreground_grid = Grid.new
        @background_grid = Grid.new
        # Initialize state
        @state = State.default
        @captions = []
        @now = Timecode.new(0, fps)
        @data_channel = 0

        update_active_grid

        magic = io.readline.chomp!
        fail InvalidFormatError, 'File does not start with "' + Scc::FILE_MAGIC + '"' unless Scc::FILE_MAGIC == magic

        io.each_line do |line|
          line.chomp!
          # Skip empty lines between the commands
          next if line.empty?

          line_data = LINE_REGEXP.match(line)
          fail InvalidFormatError, "Invalid line \"#{line}\"" if line_data.nil?

          # Parse timecode
          old_time = @now
          timecode = Timecode.new(line_data[:timecode], fps)
          @now = timecode
          fail InvalidFormatError, 'New timecode is behind last time' if @now < old_time

          # Parse data words
          parse_data(line_data[:data], check_parity)
        end
      end

      private

      # Parse one line of SCC data
      def parse_data(data, check_parity)
        last_command = [0, 0]

        data.split(' ').each do |word_string|
          begin
            @state.start_new_frame

            # Decode hexadecimal word into two-byte string
            word = [word_string].pack('H*')
            # Check parity
            fail ParityError, "At least one byte in word #{word_string} has even parity, odd required" unless !check_parity || (correct_parity?(word[0]) && correct_parity?(word[1]))

            # Remove parity bit for further processing
            word = word.bytes.collect { |byte|
              # Unset 8th bit
              (byte & ~(1 << 7))
            }

            hi, lo = word

            # First check if the word contains characters only
            if hi >= 0x20 && hi <= 0x7f
              # Skip characters if last command was on different channel
              if @data_channel != 0
                puts 'Skipping characters on channel 2'
                next
              end

              [hi, lo].each do |byte|
                handle_character(byte)
              end

              # Reset last command
              last_command = [0, 0]
            else
              if word == last_command
                # Skip commands transmitted twice for redundancy
                # But don't skip the next time, too
                last_command = [0, 0]
                next
              end

              # Channel information is encoded in the 4th bit, read it out
              @data_channel = (hi >> 3) & 1
              if @data_channel != 0
                puts 'Skipping command on channel 2'
                next
                # If channel 2 processing is needed, parse the file two times and
                # change the above condition as needed, then unset the channel bit
                # for further processing.
              end

              # rubocop:disable Style/NumericPredicate

              if hi == 0x11 && (0x30..0x3f).cover?(lo)
                # Special character
                handle_special_character(lo)
              elsif (0x12..0x13).cover?(hi) && (0x20..0x3f).cover?(lo)
                # Extended character
                handle_extended_character(hi & 1, lo)
              elsif (0x10..0x17).cover?(hi) && (0x40..0xff).cover?(lo)
                # Preamble address code
                handle_preamble_address_code(hi, lo)
              elsif [0x14, 0x17].include?(hi) && (0x20..0x2f).cover?(lo)
                handle_control_code(hi, lo)
              elsif hi == 0x11 && (0x20..0x2f).cover?(lo)
                handle_mid_row_code(hi, lo)
              elsif hi == 0x00 && lo == 0x00
                # Ignore filler
              else
                puts "Ignoring unknown command #{hi}/#{lo}"
              end

              # rubocop:enable Style/NumericPredicate

              last_command = word
            end

            post_frame if @state.paint_on_mode?
          ensure
            # Advance one frame for each word read
            @now += 1
          end
        end
      end

      # Insert one unicode character into the grid at the current position and with the
      # current style, then advance the cursor one column
      def insert_character(char)
        @active_grid[@state.row][@state.column] = Character.new(char, @state.style.dup)
        @state.column += 1
      end

      # Insert a CEA608 character into the grid at the current position, converting it to its unicode representation
      def handle_character(byte)
        # Ignore filler character
        return if byte.zero?

        char = STANDARD_CHARACTER_MAP[byte.chr]
        insert_character(char)
      end

      # Insert a special character into the grid at the current position, or delete the current column
      # in case of a transparent space.
      def handle_special_character(byte)
        if byte == 0x39
          # Transparent space: Move cursor after deleting the current column to open up a hole
          @active_grid[@state.row][@state.column] = nil
          @state.column += 1
        else
          char = SPECIAL_CHARACTER_MAP.fetch(byte)
          insert_character(char)
        end
      end

      # Insert an extended character into the grid at the current position, converting it to its Unicode representation
      def handle_extended_character(map, byte)
        char = EXTENDED_CHARACTER_MAPS.fetch(map).fetch(byte)
        # Extended characters include automatic backspace+overwrite
        @state.column -= 1
        insert_character(char)
        @state.char_replaced = true
      end

      # Set drawing position and style according to the information in a preamble address code
      def handle_preamble_address_code(hi, lo)
        @state.row = PREAMBLE_ADDRESS_CODE_ROW_MAP.fetch(hi)
        # Low byte bit 5 adds 1 to the row number if set
        @state.row += 1 if lo & (1 << 5) != 0

        # Low byte bit 0 indicates whether underlining is to be enabled
        @state.style.underline = ((lo & 1) == 1)
        # Low byte bit 4 indicates whether it is an indent or a formatting code
        is_indent = (((lo >> 4) & 1) == 1)
        # Low byte bits 1 to 3 are the color or indent code, depending on is_indent
        color_or_indent = (lo >> 1) & 0x7

        # Reset style
        @state.style.flash = false
        @state.style.italics = false

        if is_indent
          # Indent code always sets white as color attribute
          @state.style.color = Color::WHITE
          # One indent equals 4 characters
          @state.column = color_or_indent * 4
        else
          # Style code always sets first column
          @state.column = 0
          if color_or_indent == 7
            # "color" 7 is white with italics
            @state.style.color = Color::WHITE
            @state.style.italics = true
          else
            @state.style.color = Color.for_value(color_or_indent)
          end
        end
      end

      # Process a miscellaneous control code
      def handle_control_code(hi, lo)
        if hi == 0x14 && lo == 0x20
          # Resume caption loading
          @state.mode = :pop_on
          update_active_grid
        elsif hi == 0x14 && lo == 0x21
          # Backspace
          unless @state.column.zero? # Ignore in the first column
            @state.column -= 1
            # Delete character at cursor after moving one character back
            @active_grid[@state.row][@state.column] = nil
          end
        elsif hi == 0x14 && lo == 0x24
          # Delete to end of row
          (@state.column...GRID_COLUMNS).each do |column|
            @active_grid[@state.row][column] = nil
          end
        elsif hi == 0x14 && lo == 0x28
          # Flash on
          # Flash is a spacing character
          insert_character(' ')
          @state.style.flash = true
        elsif hi == 0x14 && lo == 0x29
          # Resume direct captioning
          @state.mode = :paint_on
          update_active_grid
          # elsif hi == 0x14 && lo == 0x2b
          # Resume text display -> command not described in spec
          # fail "RTD"
        elsif hi == 0x14 && lo == 0x2c
          # Erase displayed memory
          @foreground_grid = Grid.new
          post_frame
          update_active_grid
        elsif hi == 0x14 && lo == 0x2e
          # Erase non-displayed memory
          @background_grid = Grid.new
          update_active_grid
        elsif hi == 0x14 && lo == 0x2f
          # End of caption (flip memories)
          @foreground_grid, @background_grid = @background_grid, @foreground_grid
          # This also forces pop-on mode
          @state.mode = :pop_on
          post_frame

          update_active_grid
        elsif hi == 0x17 && lo >= 0x21 && lo <= 0x23
          # Tab offset
          # Bits 0 and 1 designate how many columns to go
          @state.column += (lo & 0x3)
        else
          puts "Ignoring unknown control code #{hi}/#{lo}"
        end
      end

      # Process a mid-row code
      def handle_mid_row_code(_hi, lo)
        # Mid-row codes are spacing characters
        insert_character(' ')
        # Low byte bit 0 indicates whether underlining is to be enabled
        @state.style.underline = ((lo & 1) == 1)
        # Low byte bits 1 to 3 are the color code
        color = (lo >> 1) & 0x7

        if color == 0x7
          @state.style.italics = true
        else
          # Color mid-row codes disable italics
          @state.style.italics = false
          @state.style.color = Color.for_value(color)
        end
        # All mid-row codes always disable flash
        @state.style.flash = false
      end

      # Insert the currently displayed foreground grid as caption into the captions array
      # Must be called whenever the foreground grid is changed as a result of a command
      def post_frame
        # Only push a new caption if the grid has changed, but do not push out an empty grid initially
        return unless @foreground_grid != @last_grid && !(@last_grid.nil? && @foreground_grid.empty?)

        # Save space by not saving the grid if it is completely empty
        grid = @foreground_grid.empty? ? nil : @foreground_grid
        @captions.push(Caption.new(timecode: @now, grid: grid.clone, mode: @state.mode, char_replacement: @state.char_replaced))
        @last_grid = @foreground_grid.clone
      end

      def update_active_grid
        @active_grid = @state.paint_on_mode? ? @foreground_grid : @background_grid
      end

      # Check a byte for odd parity
      def correct_parity?(byte)
        byte.ord.to_s(2).count('1').odd?
      end
    end
  end
end
